from sklearn.cluster import DBSCAN
import scanpy as sc
import anndata as ad

def dbscan_filter(adata_tmp, eps=50, min_samples=5):  
    """
    Perform DBSCAN clustering on spatial coordinates in an AnnData object.

    Parameters
    ----------
    adata_tmp
        The AnnData object containing spatial coordinates in `obsm['spatial']`.
    eps
        The maximum distance between two samples for one to be considered 
        as in the neighborhood of the other. Default is 50.
    min_samples
        The number of samples in a neighborhood for a point to be considered 
        as a core point. Default is 5.

    Returns
    -------
    AnnData
        The updated AnnData object with a new column `Bcell_aggregate_index` 
        in `obs`, containing the DBSCAN cluster labels.
        A visualization of the clustering result is also generated.

    """
    # Extract spatial coordinates for clustering
    filtered_positions = adata_tmp.obsm['spatial']
    
    # Perform DBSCAN clustering
    dbscan = DBSCAN(eps=eps, min_samples=min_samples)
    labels = dbscan.fit_predict(filtered_positions)
    
    # Add DBSCAN cluster labels to the AnnData object
    adata_tmp.obs['Bcell_aggregate_index'] = labels
    
    # Visualize the clustering results
    sc.pl.spatial(adata_tmp, color='Bcell_aggregate_index', spot_size=50, cmap='tab20_r', legend_loc='on data')
    
    # Return the updated AnnData object with clustering labels
    return adata_tmp


import numpy as np
import pandas as pd
from torchgmm.bayes import GaussianMixture as TorchGaussianMixture
from torchgmm.bayes.gmm.lightning_module import GaussianMixtureLightningModule
from torchgmm.bayes.gmm.model import GaussianMixtureModel
import logging
from sklearn.metrics import silhouette_score
from sklearn.mixture import GaussianMixture
from pytorch_lightning import Trainer
from tqdm import tqdm
import torch
from sklearn.metrics import mean_absolute_percentage_error
import anndata as ad
from copy import deepcopy
import scipy.sparse as sps
from typing import List, Tuple, cast, Union
AnyRandom = Union[int, np.random.RandomState, None]
from torchgmm.base.data import (
    DataLoader,
    TensorLike,
    collate_tensor,
    dataset_from_tensors,
)


class GaussianMixture(TorchGaussianMixture):

    model_: GaussianMixtureModel
    #: The average per-datapoint negative log-likelihood at the last training step.

    def __init__(
        self,
        n_clusters: int = 1,
        covariance_type: str = "full",
        init_strategy: str = "kmeans",
        init_means: torch.Tensor = None,
        convergence_tolerance: float = 0.001,
        covariance_regularization: float = 1e-06,
        batch_size: int = None,
        trainer_params: dict = None,
        random_state: AnyRandom = 0, #* 注意这个random_state
    ):
        super().__init__(
            num_components=n_clusters,
            covariance_type=covariance_type,
            init_strategy=init_strategy,
            init_means=init_means,
            convergence_tolerance=convergence_tolerance,
            covariance_regularization=covariance_regularization,
            batch_size=batch_size,
            trainer_params=trainer_params,
        )
        self.n_clusters = n_clusters
        self.random_state = random_state

    def score(self, data: TensorLike):
        return super().score(data) 

    
    def fit(self, data: TensorLike):
        if sps.issparse(data):
            raise ValueError(
                "Sparse data is not supported. You may have forgotten to reduce the dimensionality of the data. Otherwise, please convert the data to a dense format."
            )
        return self._fit(data)

    def _fit(self, data):
        try:
            return super().fit(data)
        except torch._C._LinAlgError as e:
            if self.covariance_regularization >= 1:
                raise ValueError(
                    "Cholesky decomposition failed even with covariance regularization = 1. The matrix may be singular."
                ) from e
            else:
                self.covariance_regularization *= 10
                logger.warning(
                    f"Cholesky decomposition failed. Retrying with covariance regularization {self.covariance_regularization}."
                )
                return self._fit(data)

    def predict(self, data: TensorLike):
        """ 
        Computes the most likely components for each of the provided datapoints.
        
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        return super().predict(data).numpy()

    def predict_proba(self, data: TensorLike):
        """
        Computes a distribution over the components for each of the provided datapoints.
        
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        loader = DataLoader(
            dataset_from_tensors(data),
            batch_size=self.batch_size or len(data),
            collate_fn=collate_tensor,
        ) 
        trainer_params = self.trainer_params.copy()
        trainer_params["logger"] = False
        result = Trainer(**trainer_params).predict(GaussianMixtureLightningModule(self.model_), loader)
        return torch.cat([x[0] for x in cast(List[Tuple[torch.Tensor, torch.Tensor]], result)])

    def score_samples(self, data: TensorLike):
        """
        Computes the negative log-likelihood (NLL) of each of the provided datapoints.
        
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        loader = DataLoader(
            dataset_from_tensors(data),
            batch_size=self.batch_size or len(data),
            collate_fn=collate_tensor,
        )
        trainer_params = self.trainer_params.copy()
        trainer_params["logger"] = False
        result = Trainer(**trainer_params).predict(GaussianMixtureLightningModule(self.model_), loader)
        return torch.stack([x[1] for x in cast(List[Tuple[torch.Tensor, torch.Tensor]], result)])




class ClusterAutoK:
    """
    Identify the best candidates for the number of clusters.
    """
    silhouette_scores: np.ndarray

    def __init__(
        self,
        n_clusters,
        max_runs: int = 5,
        convergence_tol: float = 1e-2,
        model_class: type = None,
    ):
        self.n_clusters = (
            list(range(*(max(2, n_clusters[0] - 1), n_clusters[1] + 1)))
            if isinstance(n_clusters, tuple)
            else n_clusters
        )
        self.max_runs = max_runs
        self.convergence_tol = convergence_tol
        self.model_class = model_class if model_class else GaussianMixture
        self.silhouette_scores = []

    def fit(self, adata: ad.AnnData, use_rep: str = "X_cellcharter"): 
        if use_rep not in adata.obsm:
            raise ValueError(f"{use_rep} not found in adata.obsm. If you want to use adata.X, set use_rep=None") 

        X = adata.obsm[use_rep] if use_rep is not None else adata.X 
        
        self.best_models = {} 
        random_state = 0
        previous_silhouette_scores = None
        for i in range(self.max_runs):
            print(f"Iteration {i + 1}/{self.max_runs}")
            run_silhouette_scores = []

            for k in tqdm(self.n_clusters, disable=(len(self.n_clusters) == 1)):
                # Suppress PyTorch warnings
                logging_level = logging.getLogger("lightning.pytorch").getEffectiveLevel()
                logging.getLogger("lightning.pytorch").setLevel(logging.ERROR)
                clustering = self.model_class(n_clusters=k, random_state=i + random_state)
                clustering.fit(X)  
                logging.getLogger("lightning.pytorch").setLevel(logging_level)

                # Track best models by negative log-likelihood (nll_)
                if (k not in self.best_models.keys()) or (clustering.nll_ < self.best_models[k].nll_):
                    self.best_models[k] = clustering

                silhouette_avg = silhouette_score(X, clustering.predict(X))
                run_silhouette_scores.append(silhouette_avg)
        
            self.silhouette_scores.append(run_silhouette_scores) 
            print(self.silhouette_scores)
            
            if i > 0:
                if previous_silhouette_scores is not None:
                    silhouette_scores_change = mean_absolute_percentage_error(
                        np.mean(previous_silhouette_scores, axis=0),
                        np.mean(self.silhouette_scores, axis=0),
                    )
                    if silhouette_scores_change < self.convergence_tol:
                        print(f"Convergence with a change in silhouette_scores of {silhouette_scores_change} reached after {i + 1} iterations")
                        break 
                previous_silhouette_scores = deepcopy(self.silhouette_scores)

    @property
    def best_k(self) -> int:
        """The number of clusters with the highest silhouette_scores."""
        if self.max_runs <= 1:
            raise ValueError("Cannot compute silhouette_scores with max_runs <= 1")
        silhouette_scores_mean = np.mean(self.silhouette_scores, axis=0) 
        best_idx = np.argmax(silhouette_scores_mean)
        return self.n_clusters[best_idx]

    def predict(self, adata: ad.AnnData, use_rep: str = None, k: int = None) -> pd.Categorical:
        """
        Predict the labels for the data in ``use_rep`` using the fitted model.
        """
        k = self.best_k if k is None else k
        assert k is None or k in self.n_clusters

        X = (
            adata.obsm[use_rep]
            if use_rep is not None
            else adata.obsm["X_cellcharter"] if "X_cellcharter" in adata.obsm else adata.X
        )
        return pd.Categorical(self.best_models[k].predict(X), categories=np.arange(k))



# import scanpy as sc
# import matplotlib.pyplot as plt

# def enrichment_score(adata_tmp, score_name, if_filtered=False):
#     """
#     Plot the spatial data, either the original or filtered based on the `if_filtered` parameter.
    
#     Parameters
#     ----------
#     adata_tmp: AnnData object
#         The spatial data.
#     score_name: str
#         the name of the score (e.g., Bcell_enrichment) to color the plot.
#     if_filtered: bool
#         whether to plot the filtered data (True) or the original data (False).
#     """
#     fig, ax = plt.subplots(figsize=(4.5, 4.5))
    
#     if if_filtered:
#         # Plot the filtered spatial data
#         if 'filtered_coords' in adata_tmp.obs:
#             filtered_adata = adata_tmp[adata_tmp.obs['filtered_coords']]
#             sc.pl.spatial(filtered_adata, color=score_name, ax=ax, show=False, spot_size=50)
#         else:
#             print("Warning: 'filtered_coords' not found. Plotting original data.")
#             sc.pl.spatial(adata_tmp, color=score_name, ax=ax, show=False, spot_size=50)
#     else:
#         # Plot the original spatial data
#         sc.pl.spatial(adata_tmp, color=score_name, ax=ax, show=False, spot_size=50)
    
#     plt.show()



import logging
import torch
from torchgmm.bayes import GaussianMixture as TorchGaussianMixture
from torchgmm.bayes.gmm.lightning_module import GaussianMixtureLightningModule
from torchgmm.bayes.gmm.model import GaussianMixtureModel
from torchgmm.base.data import (
    DataLoader,
    TensorLike,
    collate_tensor,
    dataset_from_tensors,
)

from pytorch_lightning import Trainer
from typing import List, Tuple, cast, Union
from tqdm.auto import tqdm
from scipy.sparse import spdiags
import anndata as ad
import numpy as np
import pandas as pd
import scipy.sparse as sps


AnyRandom = Union[int, np.random.RandomState, None]

class GaussianMixture(TorchGaussianMixture):
    """
    Adapted version of GaussianMixture clustering model from the `torchgmm <https://github.com/marcovarrone/torchgmm/>`_ library.

    Parameters
    ----------
    n_clusters
        The number of components in the GMM. The dimensionality of each component is automatically inferred from the data.
    covariance_type
        The type of covariance to assume for all Gaussian components.
    init_strategy
        The strategy for initializing component means and covariances.
    init_means
        An optional initial guess for the means of the components. If provided,
        must be a tensor of shape ``[num_components, num_features]``. If this is given,
        the ``init_strategy`` is ignored and the means are handled as if K-means
        initialization has been run.
    convergence_tolerance
        The change in the per-datapoint negative log-likelihood which
        implies that training has converged.
    covariance_regularization
        A small value which is added to the diagonal of the
        covariance matrix to ensure that it is positive semi-definite.
    batch_size: The batch size to use when fitting the model. If not provided, the full
        data will be used as a single batch. Set this if the full data does not fit into
        memory.
    trainer_params
        Initialization parameters to use when initializing a PyTorch Lightning
        trainer. By default, it disables various stdout logs unless TorchGMM is configured to
        do verbose logging. Checkpointing and logging are disabled regardless of the log
        level. This estimator further sets the following overridable defaults:
        - ``max_epochs=100``.
    random_state
        Initialization seed.

    """

    model_: GaussianMixtureModel
    #: The average per-datapoint negative log-likelihood at the last training step.

    def __init__(
        self,
        n_clusters: int = 1,
        covariance_type: str = "full",
        init_strategy: str = "kmeans",
        init_means: torch.Tensor = None,
        convergence_tolerance: float = 0.001,
        covariance_regularization: float = 1e-06,
        batch_size: int = None,
        trainer_params: dict = None,
        random_state: AnyRandom = 0,
    ):
        super().__init__(
            num_components=n_clusters,
            covariance_type=covariance_type,
            init_strategy=init_strategy,
            init_means=init_means,
            convergence_tolerance=convergence_tolerance,
            covariance_regularization=covariance_regularization,
            batch_size=batch_size,
            trainer_params=trainer_params,
        )
        self.n_clusters = n_clusters
        self.random_state = random_state

    def score(self, data: TensorLike):
        return super().score(data) 

    
    def fit(self, data: TensorLike):
        if sps.issparse(data):
            raise ValueError(
                "Sparse data is not supported. You may have forgotten to reduce the dimensionality of the data. Otherwise, please convert the data to a dense format."
            )
        return self._fit(data)

    def _fit(self, data):
        try:
            return super().fit(data)
        except torch._C._LinAlgError as e:
            if self.covariance_regularization >= 1:
                raise ValueError(
                    "Cholesky decomposition failed even with covariance regularization = 1. The matrix may be singular."
                ) from e
            else:
                self.covariance_regularization *= 10
                logger.warning(
                    f"Cholesky decomposition failed. Retrying with covariance regularization {self.covariance_regularization}."
                )
                return self._fit(data)

    def predict(self, data: TensorLike):
        """ 
        Computes the most likely components for each of the provided datapoints.

        Parameters
        ----------
        data
            The datapoints for which to obtain the most likely components.
        Returns
        ----------
            A tensor of shape ``[num_datapoints]`` with the indices of the most likely components.
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        return super().predict(data).numpy()

    def predict_proba(self, data: TensorLike):
        """
        Computes a distribution over the components for each of the provided datapoints.
        
        Parameters
        ----------
        data
            The datapoints for which to compute the component assignment probabilities.
        Returns
        ----------
            A tensor of shape ``[num_datapoints, num_components]`` with the assignment
            probabilities for each component and datapoint. Note that each row of the vector sums
            to 1, i.e. the returned tensor provides a proper distribution over the components for
            each datapoint.
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        loader = DataLoader(
            dataset_from_tensors(data),
            batch_size=self.batch_size or len(data),
            collate_fn=collate_tensor,
        ) 
        trainer_params = self.trainer_params.copy()
        trainer_params["logger"] = False
        result = Trainer(**trainer_params).predict(GaussianMixtureLightningModule(self.model_), loader)
        return torch.cat([x[0] for x in cast(List[Tuple[torch.Tensor, torch.Tensor]], result)])

    def score_samples(self, data: TensorLike):
        """
        Computes the negative log-likelihood (NLL) of each of the provided datapoints.
        
        Parameters
        ----------
        data
            The datapoints for which to compute the NLL.
        Returns
        ----------
            A tensor of shape ``[num_datapoints]`` with the NLL for each datapoint.
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        loader = DataLoader(
            dataset_from_tensors(data),
            batch_size=self.batch_size or len(data),
            collate_fn=collate_tensor,
        )
        trainer_params = self.trainer_params.copy()
        trainer_params["logger"] = False
        result = Trainer(**trainer_params).predict(GaussianMixtureLightningModule(self.model_), loader)
        return torch.stack([x[1] for x in cast(List[Tuple[torch.Tensor, torch.Tensor]], result)])

class Cluster(GaussianMixture):
    """
    Cluster cells or spots based on features.

    Parameters
    ----------
    n_clusters
        The number of components in the GMM. The dimensionality of each component is automatically inferred from the data.
    covariance_type
        The type of covariance to assume for all Gaussian components.
    init_strategy
        The strategy for initializing component means and covariances.
    init_means
        An optional initial guess for the means of the components. If provided,
        must be a tensor of shape ``[num_components, num_features]``. If this is given,
        the ``init_strategy`` is ignored and the means are handled as if K-means
        initialization has been run.
    convergence_tolerance
        The change in the per-datapoint negative log-likelihood which
        implies that training has converged.
    covariance_regularization
        A small value which is added to the diagonal of the
        covariance matrix to ensure that it is positive semi-definite.
    batch_size: The batch size to use when fitting the model. If not provided, the full
        data will be used as a single batch. Set this if the full data does not fit into
        memory.
    trainer_params
        Initialization parameters to use when initializing a PyTorch Lightning
        trainer. By default, it disables various stdout logs unless TorchGMM is configured to
        do verbose logging. Checkpointing and logging are disabled regardless of the log
        level. This estimator further sets the following overridable defaults:
        - ``max_epochs=100``.
    random_state
        Initialization seed.
    Examples
    --------
    >>> adata = anndata.read_h5ad(path_to_anndata)
    >>> sq.gr.spatial_neighbors(adata, coord_type='generic', delaunay=True)
    >>> cc.gr.remove_long_links(adata)
    >>> cc.gr.aggregate_neighbors(adata, n_layers=3)
    >>> model = cc.tl.Cluster(n_clusters=11)
    >>> model.fit(adata, use_rep='X_cellcharter')
    """

    def __init__(
        self,
        n_clusters: int = 1, 
        covariance_type: str = "full",
        init_strategy: str = "kmeans",
        init_means: torch.Tensor = None,
        convergence_tolerance: float = 0.001,
        covariance_regularization: float = 1e-06,
        batch_size: int = None,
        trainer_params: dict = None,
        random_state: AnyRandom = 0,
    ):
        super().__init__(                            # 这个调用将当前类的初始化参数传递给父类的构造函数（super().__init__()）
            n_clusters=n_clusters,
            covariance_type=covariance_type,
            init_strategy=init_strategy,
            init_means=init_means,
            convergence_tolerance=convergence_tolerance,
            covariance_regularization=covariance_regularization,
            batch_size=batch_size,
            trainer_params=trainer_params,
            random_state=random_state,
        )

    def fit(self, adata: ad.AnnData, use_rep: str = "X_cellcharter"): #* 前面aggragate的结果就是X_cellcharter
        """
        Fit data into ``n_clusters`` clusters.

        Parameters
        ----------
        adata
            Annotated data object.
        use_rep
            Key in :attr:`anndata.AnnData.obsm` to use as data to fit the clustering model.
        """
        logging_level = logging.root.level 
        X = adata.X if use_rep is None else adata.obsm[use_rep]
        logging_level = logging.getLogger("lightning.pytorch").getEffectiveLevel()
        
        logging.getLogger("lightning.pytorch").setLevel(logging.ERROR)
        super().fit(X)
        logging.getLogger("lightning.pytorch").setLevel(logging_level) 
        # 临时忽略低于 ERROR 级别的日志（例如，WARNING、INFO、DEBUG 等信息将不再输出）
        
        adata.uns["_cellcharter"] = {k: v for k, v in self.get_params().items() if k != "init_means"}
    
    def score(self, adata: ad.AnnData, use_rep: str = "X_cellcharter"):
        """
        Fit data into ``n_clusters`` clusters.
        """ 
        logging_level = logging.root.level 
        X =  adata.obsm[use_rep]
        logging_level = logging.getLogger("lightning.pytorch").getEffectiveLevel()
        
        logging.getLogger("lightning.pytorch").setLevel(logging.ERROR)
        super().score(X)
        
        logging.getLogger("lightning.pytorch").setLevel(logging_level) 
        # 临时忽略低于 ERROR 级别的日志（例如，WARNING、INFO、DEBUG 等信息将不再输出）
        
        adata.uns["_cellcharter"] = {k: v for k, v in self.get_params().items() if k != "init_means"}

    def predict(self,
            adata: ad.AnnData,
            use_rep: str = "X_cellcharter",
            store_labels: bool = False,  
            store_column: str = 'predicted_labels'):  
        """
        Predict the labels for the data in ``use_rep`` using the fitted model.
        
        Parameters
        ----------
        adata
            AnnData object containing the data to predict.
        use_rep
            str, the name of the key to use from `adata.obsm` or `adata.X` (default is 'X_cellcharter').
        store_labels
            bool, whether to store the predicted labels in `adata.obs` (default is False).
        store_column
            str, the name of the column to store the predicted labels (default is 'predicted_labels').
        
        Returns
        ----------
            A pandas Categorical object containing the predicted labels.
        """
        X = adata.X if use_rep is None else adata.obsm[use_rep]
        
        labels = pd.Categorical(super().predict(X), categories=np.arange(self.n_clusters))
        
        if store_labels:
            adata.obs[store_column] = labels 
    
        return labels




import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import anndata as ad

def clustering_results(adata: ad.AnnData, cluster_column: str, color_dict: dict = None, spot_size: int = 50):
    """
    Plots the clustering results, displaying the cluster labels for each data point using the specified colors.
    
    Parameters
    ----------
    adata: AnnData object
        containing spatial information and clustering labels.
    - cluster_column: str
        the name of the column containing the cluster labels.(eg. the name of the column in use_rep in cluster_Auto_k)
    - color_dict: dict
        a dictionary specifying the color for each cluster label (optional).
    - spot_size: int
        the size of the data points (default is 50).
    """
    color_lst = [ 
    '#C9E9D9', # 0
    '#ffc0cb', # 1
    '#FEFDD3', # 2 #F8E2E1  ok 
    '#C1B1D2', # 3 ok 
    '#EEE5F8', # 4 ok
    '#4074AA', # 5 ok 
    '#EF9194', # 6 ok 
    '#D0353A', # 7 ok
    '#78A5C9', # 8 ok
    '#798af2', # 9 ok
    '#E4F3FC', # 10  cceeff ok
    '#8BD0D8', # 11 ok
    '#F8E886', # 12 f5cfdc
    '#A4C4D9'
]

    color_dict = {}
    cluster_lst = np.sort(list(set(adata.obs[cluster_column])))
    for i in range(len(cluster_lst)): 
        color_dict[cluster_lst[i]] = color_lst[i]
    # Create the plot
    fig, ax = plt.subplots(figsize=(8, 8))

    # Use Scanpy's spatial plotting function
    sc.pl.spatial(adata, color=[cluster_column], ax=ax, show=False, spot_size=spot_size, palette=color_lst)
    
    # Display the plot
    plt.title(f"Clustering results for {cluster_column}")  #这里的title可以改一下
    plt.show()



from typing import List, Tuple, cast, Union
from tqdm.auto import tqdm
from scipy.sparse import spdiags
import anndata as ad
import numpy as np
import pandas as pd
import scipy.sparse as sps
import warnings
warnings.filterwarnings('ignore') 


def str2list(value: Union[str, list]) -> list:
    """Check whether value is a string. If so, converts into a list containing value."""
    return [value] if isinstance(value, str) else value


def _hop(adj_hop, adj, adj_visited=None):
    adj_hop = adj_hop @ adj    

    if adj_visited is not None:
        adj_hop = adj_hop > adj_visited  # Logical not for sparse matrices
        adj_visited = adj_visited + adj_hop
    return adj_hop, adj_visited


def _mul_broadcast(mat1, mat2):
    return spdiags(mat2, 0, len(mat2), len(mat2)) * mat1
    
def _normalize(adj):
    deg = np.array(np.sum(adj, axis=1)).squeeze()
    with warnings.catch_warnings():
        # If a cell doesn't have neighbors deg = 0 -> divide by zero
        warnings.filterwarnings(action="ignore", category=RuntimeWarning)
        deg_inv = 1 / deg
    deg_inv[deg_inv == float("inf")] = 0
    return _mul_broadcast(adj, deg_inv) 

def _setdiag(array, value):  
    if isinstance(array, sps.csr_matrix):
        array = array.tolil()
    array.setdiag(value)
    array = array.tocsr()
    if value == 0:
        array.eliminate_zeros()
    return array

def _aggregate(adj, x, method):
    if method == "mean":
        return _aggregate_mean(adj, x)
    elif method == "var":
        return _aggregate_var(adj, x)
    else:
        raise NotImplementedError
###
def _aggregate_mean(adj, x):  
    return adj @ x

def _aggregate_var(adj, x):
    mean = adj @ x
    mean_squared = adj @ (x * x)
    return mean_squared - mean * mean

def _aggregate(adj, x, method):
    if method == "mean":
        return _aggregate_mean(adj, x)
    elif method == "var":
        return _aggregate_var(adj, x)
    else:
        raise NotImplementedError
###
def _aggregate_neighbors(
    adj: sps.spmatrix,
    X: np.ndarray,
    nhood_layers: list,
    aggregations:  "mean",
    disable_tqdm: bool = True,
) -> np.ndarray:
    adj = adj.astype(bool)
    adj = _setdiag(adj, 0)
    adj_hop = adj.copy() 
    adj_visited = _setdiag(adj.copy(), 1) 


    Xs = []
    for i in tqdm(range(0, max(nhood_layers) + 1), disable=disable_tqdm):
        if i in nhood_layers:
            if i == 0:
                Xs.append(X)
            else:
                if i > 1:
                    adj_hop, adj_visited = _hop(adj_hop, adj, adj_visited)
                adj_hop_norm = _normalize(adj_hop) 
                
                for agg in aggregations:
                    x = _aggregate(adj_hop_norm, X, agg)
                    
                    Xs.append(x)
    if sps.issparse(X):
        return sps.hstack(Xs)
    else:
        return np.hstack(Xs)


def aggregate_neighbors(
    adata,
    n_layers, 
    aggregations =  "mean",
    connectivity_key  = None,  
    use_rep  = None, 
    sample_key = None,
    out_key  = "X_cellcharter",
    copy: bool = False,  
) :
    """
    Aggregate the features from each neighborhood layers and concatenate them
    """
    connectivity_key = connectivity_key
    sample_key = Key.obs.sample if sample_key is None else sample_key
    aggregations = str2list(aggregations)

    X = adata.X if use_rep is None else adata.obsm[use_rep]

    if isinstance(n_layers, int):
        n_layers = list(range(n_layers + 1))
    
    if sps.issparse(X):
        X_aggregated = sps.dok_matrix(
            (X.shape[0], X.shape[1] * ((len(n_layers) - 1) * len(aggregations) + 1)), dtype=np.float32     
        )
    else:
        X_aggregated = np.empty(
            (X.shape[0], X.shape[1] * ((len(n_layers) - 1) * len(aggregations) + 1)), dtype=np.float32
        )

    if sample_key in adata.obs:
        samples = adata.obs[sample_key].unique()    
        sample_idxs = [adata.obs[sample_key] == sample for sample in samples]
    else:
        sample_idxs = [np.arange(adata.shape[0])]

    for idxs in tqdm(sample_idxs, disable=(len(sample_idxs) == 1)):
        X_sample_aggregated = _aggregate_neighbors(
            adj=adata[idxs].obsp[connectivity_key],
            X=X[idxs],
            nhood_layers=n_layers,
            aggregations=aggregations,
            disable_tqdm=(len(sample_idxs) != 1),
        )
        X_aggregated[idxs] = X_sample_aggregated

    if isinstance(X_aggregated, sps.dok_matrix):
        X_aggregated = X_aggregated.tocsr()

    if copy:
        return X_aggregated

    adata.obsm[out_key] = X_aggregated


import sys
# module_2.py
sys.path.append(os.path.join(os.path.dirname(__file__), '../B_HIT'))
from tl import calculate_high_density_regions
from pl import enrichment_score
import scanpy as sc


# 读取原始数据
adata = sc.read('/home/yuanrh/cancerdata/onesample.h5ad')
signature = ["CD79A", "CD79B", "MS4A1", "CD79A", 'CD79B', "MZB1", "JCHAIN", "IGHA1", 'IGHG1', 'IGHG3']
score_name = 'Bcell_enrichment'
high_binSize = 100
sc.tl.score_genes(adata, gene_list=signature, score_name=score_name)
adata = adata[adata.obs[score_name]>0] 
calculate_high_density_regions(adata, score_name)
enrichment_score(adata, score_name=score_name, if_filtered=True)


def get_adj(adata, mode = 'neighbour', k_neighbors = 7, min_distance = 150, self_loop = True):
    """
    Calculate adjacency matrix for ST data.
    
    Parameters
    ----------
    mode : str ['neighbour','distance'] (default: 'neighbour')
        The way to define neighbourhood. 
        If `mode='neighbour'`: Calculate adjacency matrix with specified number of nearest neighbors;
        If `mode='distance'`: Calculate adjacency matrix with neighbors within the specified distance.
    k_neighbors : int (default: 7)
        For `mode = 'neighbour'`, set the number of nearest neighbors if `mode='neighbour'`.
    min_distance : int (default: 150)
        For `mode = 'distance'`, set the distance of nearest neighbors if `mode='distance'`.
    self_loop : bool (default: True)
        Whether to add selfloop to the adjacency matrix.
        
    adj : matrix of shape (n_samples, n_samples)
        Adjacency matrix where adj[i, j] is assigned the weight of edge that connects i to j.
    """
    spatial = adata.obsm["spatial"]   
    if mode == 'distance':
        assert min_distance is not None,"Please set `min_diatance` for `get_adj()`"
        adj = metrics.pairwise_distances(spatial, metric='euclidean')
        adj[adj > min_distance] = 0
        if self_loop:
            adj += np.eye(adj.shape[0])  
        adj = np.int64(adj>0)
        return adj
    
    elif mode == 'neighbour':
        assert k_neighbors is not None,"Please set `k_neighbors` for `get_adj()`"
        adj = kneighbors_graph(spatial, n_neighbors = k_neighbors, include_self = self_loop)
        return adj


import logging
import torch
from torchgmm.bayes import GaussianMixture as TorchGaussianMixture
from torchgmm.bayes.gmm.lightning_module import GaussianMixtureLightningModule
from torchgmm.bayes.gmm.model import GaussianMixtureModel
from torchgmm.base.data import (
    DataLoader,
    TensorLike,
    collate_tensor,
    dataset_from_tensors,
)

from pytorch_lightning import Trainer
from typing import List, Tuple, cast, Union
from tqdm.auto import tqdm
from scipy.sparse import spdiags
import anndata as ad
import numpy as np
import pandas as pd
import scipy.sparse as sps


AnyRandom = Union[int, np.random.RandomState, None]

class GaussianMixture(TorchGaussianMixture):

    model_: GaussianMixtureModel
    #: The average per-datapoint negative log-likelihood at the last training step.

    def __init__(
        self,
        n_clusters: int = 1,
        covariance_type: str = "full",
        init_strategy: str = "kmeans",
        init_means: torch.Tensor = None,
        convergence_tolerance: float = 0.001,
        covariance_regularization: float = 1e-06,
        batch_size: int = None,
        trainer_params: dict = None,
        random_state: AnyRandom = 0,
    ):
        super().__init__(
            num_components=n_clusters,
            covariance_type=covariance_type,
            init_strategy=init_strategy,
            init_means=init_means,
            convergence_tolerance=convergence_tolerance,
            covariance_regularization=covariance_regularization,
            batch_size=batch_size,
            trainer_params=trainer_params,
        )
        self.n_clusters = n_clusters
        self.random_state = random_state

    def score(self, data: TensorLike):
        return super().score(data) 

    
    def fit(self, data: TensorLike):
        if sps.issparse(data):
            raise ValueError(
                "Sparse data is not supported. You may have forgotten to reduce the dimensionality of the data. Otherwise, please convert the data to a dense format."
            )
        return self._fit(data)

    def _fit(self, data):
        try:
            return super().fit(data)
        except torch._C._LinAlgError as e:
            if self.covariance_regularization >= 1:
                raise ValueError(
                    "Cholesky decomposition failed even with covariance regularization = 1. The matrix may be singular."
                ) from e
            else:
                self.covariance_regularization *= 10
                logger.warning(
                    f"Cholesky decomposition failed. Retrying with covariance regularization {self.covariance_regularization}."
                )
                return self._fit(data)

    def predict(self, data: TensorLike):
        """ 
        Computes the most likely components for each of the provided datapoints.
        
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        return super().predict(data).numpy()

    def predict_proba(self, data: TensorLike):
        """
        Computes a distribution over the components for each of the provided datapoints.
        
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        loader = DataLoader(
            dataset_from_tensors(data),
            batch_size=self.batch_size or len(data),
            collate_fn=collate_tensor,
        ) 
        trainer_params = self.trainer_params.copy()
        trainer_params["logger"] = False
        result = Trainer(**trainer_params).predict(GaussianMixtureLightningModule(self.model_), loader)
        return torch.cat([x[0] for x in cast(List[Tuple[torch.Tensor, torch.Tensor]], result)])

    def score_samples(self, data: TensorLike):
        """
        Computes the negative log-likelihood (NLL) of each of the provided datapoints.
        
        Attention
        ----------
            When calling this function in a multi-process environment, each process receives only
            a subset of the predictions. If you want to aggregate predictions, make sure to gather
            the values returned from this method.
        """
        loader = DataLoader(
            dataset_from_tensors(data),
            batch_size=self.batch_size or len(data),
            collate_fn=collate_tensor,
        )
        trainer_params = self.trainer_params.copy()
        trainer_params["logger"] = False
        result = Trainer(**trainer_params).predict(GaussianMixtureLightningModule(self.model_), loader)
        return torch.stack([x[1] for x in cast(List[Tuple[torch.Tensor, torch.Tensor]], result)])

class Cluster(GaussianMixture):
    """
    Cluster cells or spots based on features          ###### neighborhood aggregated features from CellCharter.
    
    """

    def __init__(
        self,
        n_clusters: int = 1, 
        covariance_type: str = "full",
        init_strategy: str = "kmeans",
        init_means: torch.Tensor = None,
        convergence_tolerance: float = 0.001,
        covariance_regularization: float = 1e-06,
        batch_size: int = None,
        trainer_params: dict = None,
        random_state: AnyRandom = 0,
    ):
        super().__init__(                            # 这个调用将当前类的初始化参数传递给父类的构造函数（super().__init__()）
            n_clusters=n_clusters,
            covariance_type=covariance_type,
            init_strategy=init_strategy,
            init_means=init_means,
            convergence_tolerance=convergence_tolerance,
            covariance_regularization=covariance_regularization,
            batch_size=batch_size,
            trainer_params=trainer_params,
            random_state=random_state,
        )

    def fit(self, adata: ad.AnnData, use_rep: str = "X_cellcharter"): #* 前面aggragate的结果就是X_cellcharter
        """
        Fit data into ``n_clusters`` clusters.
        """ 
        logging_level = logging.root.level 
        X = adata.X if use_rep is None else adata.obsm[use_rep]
        logging_level = logging.getLogger("lightning.pytorch").getEffectiveLevel()
        
        logging.getLogger("lightning.pytorch").setLevel(logging.ERROR)
        super().fit(X)
        logging.getLogger("lightning.pytorch").setLevel(logging_level) 
        # 临时忽略低于 ERROR 级别的日志（例如，WARNING、INFO、DEBUG 等信息将不再输出）
        
        adata.uns["_cellcharter"] = {k: v for k, v in self.get_params().items() if k != "init_means"}
    
    def score(self, adata: ad.AnnData, use_rep: str = "X_cellcharter"):
        """
        Fit data into ``n_clusters`` clusters.
        """ 
        logging_level = logging.root.level 
        X =  adata.obsm[use_rep]
        logging_level = logging.getLogger("lightning.pytorch").getEffectiveLevel()
        
        logging.getLogger("lightning.pytorch").setLevel(logging.ERROR)
        super().score(X)
        
        logging.getLogger("lightning.pytorch").setLevel(logging_level) 
        # 临时忽略低于 ERROR 级别的日志（例如，WARNING、INFO、DEBUG 等信息将不再输出）
        
        adata.uns["_cellcharter"] = {k: v for k, v in self.get_params().items() if k != "init_means"}

    def predict(self,
            adata: ad.AnnData,
            use_rep: str = "X_cellcharter",
            store_labels: bool = False,  # 新增参数，用于控制是否存储标签
            store_column: str = 'predicted_labels'):  # 新增参数，用于指定存储的列名
        """
        Predict the labels for the data in ``use_rep`` using the fitted model.
        
        Parameters:
        - adata: AnnData object containing the data to predict.
        - use_rep: str, the name of the key to use from `adata.obsm` or `adata.X` (default is 'X_cellcharter').
        - store_labels: bool, whether to store the predicted labels in `adata.obs` (default is False).
        - store_column: str, the name of the column to store the predicted labels (default is 'predicted_labels').
        
        Returns:
        - A pandas Categorical object containing the predicted labels.
        """
        X = adata.X if use_rep is None else adata.obsm[use_rep]
        
        labels = pd.Categorical(super().predict(X), categories=np.arange(self.n_clusters))
        
        if store_labels:
            adata.obs[store_column] = labels 
    
        return labels




from joblib import Parallel, delayed
from sklearn.neighbors import KernelDensity
from collections import defaultdict
import numpy as np
from scipy.ndimage import binary_opening
import multiprocessing
import time
import warnings
warnings.filterwarnings('ignore') 

def calculate_density(kde, pos):
    return np.exp(kde.score_samples([pos]))
    
def kde_filter(adata_tmp,
               score_name,
               high_binSize = 100,
               default_thread_num=36,
               clean_mask_size=(3,3) ):
    
    """
    Filter spatial transcriptomics (ST) data using Kernel Density Estimation (KDE).

    Parameters
    ----------
    adata_tmp : AnnData object
        The input AnnData object containing spatial and score data.
    score_name : str
        The name of the column in `adata_tmp.obs` that contains the scores to be used as weights.
    high_binSize : int, optional (default: 100)
        The size of the bins used for spatial discretization.
    default_thread_num : int, optional (default: 36)
        The maximum number of threads to use for parallel KDE scoring.
    clean_mask_size : tuple of int, optional (default: (3, 3))
        The size of the structuring element used for binary opening during mask cleaning.

    Returns
    -------
    AnnData
        A subset of the input AnnData object, filtered to retain only the valid spatial coordinates.
    
    """


    # original locations 
    start_time = time.time()
    positions_original = adata_tmp.obsm['spatial']

    # prepare
    x = np.array(positions_original[:, 0]); y = np.array(positions_original[:, 1])
    x_min, x_max = x.min(), x.max()
    y_min, y_max = y.min(), y.max()
    Z = np.array(adata_tmp.obs[score_name]).astype(np.float32)
    high_binSize = high_binSize
    bandwidth = int( high_binSize / 3 )
    points = np.vstack([x, y]).T  

    # record original coords
    xy_dict = defaultdict(list)   
    for i in range(len(x)):
        Bx = (x[i] - x_min) // high_binSize * high_binSize + x_min
        By = (y[i] - y_min) // high_binSize * high_binSize + y_min
        xy_dict[(Bx, By)].append([x[i], y[i]])  # 直接追加，无需检查键是否存在
        
    # kde fit
    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth)
    kde.fit(points, sample_weight=Z)   
    x_grid, y_grid = np.meshgrid(np.arange(x_min,x_max,high_binSize), np.arange(y_min,y_max,high_binSize)) 
    positions = np.vstack([x_grid.ravel(), y_grid.ravel()]).T     # 	np.vstack 是 np.concatenate 函数的一个方便函数，等价于 np.concatenate，但指定了 axis=0 即在第一个轴（行）上进行堆叠
    x_grid = x_grid.astype(np.float32); y_grid = y_grid.astype(np.float32); positions = positions.astype(np.float32)
    print(f"Step 1 (Prepare): {time.time() - start_time:.2f} s")
    
    # kde score
    start_time = time.time()
    num_cores = min(multiprocessing.cpu_count(), default_thread_num)
    density = Parallel(n_jobs=num_cores)(delayed(calculate_density)(kde, pos) for pos in positions) 
    density = (np.array(density).reshape(x_grid.shape))    
    print(f"Step 2 (kde score): {time.time() - start_time:.2f} s")

    # set density cutoff
    start_time = time.time()
    density_flat = density.ravel()
    # threshold = np.percentile(density_flat, 50) 
    threshold = np.mean(density_flat)                ################### Cutoff is setting as mean value  ###################

    density_thresholded = np.where(density > threshold, density, np.nan)
    binary_mask = ~np.isnan(density_thresholded)
    cleaned_mask = binary_opening(binary_mask, structure=np.ones(clean_mask_size))
    density_cleaned = np.where(cleaned_mask, density, np.nan) 
    mask = ~np.isnan(density_cleaned)
    x_coords = x_grid[mask]
    y_coords = y_grid[mask]

    adata_tmp.uns["binary_mask"] = binary_mask
    adata_tmp.uns["mask"] = mask
    
    coords = np.vstack([x_coords, y_coords]).T
    orginal_coords = []
    for coord in coords:
        orginal_coords += (xy_dict[tuple(coord)])

    # obtain the legal coords
    original_coords_set = {tuple(row) for row in orginal_coords}
    isin_ = [tuple(x) in original_coords_set for x in positions_original]
    print(f"Step 3 (obtain legal coords): {time.time() - start_time:.2f} s")

    adata_tmp.obs['filtered_coords'] = isin_

    return adata_tmp[isin_]


import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

def silhouette_scores(model, k_range: tuple, figsize=(6, 4)):
    """
    Plot silhouette scores for a given range of clusters.
    
    Parameters
    ----------
    model: object
        The clustering model that has an attribute `silhouette_scores` (list of lists with silhouette scores).
        
    k_range: tuple
        A tuple (n, m) representing the range of number of clusters to evaluate.
        
    figsize: tuple, optional
        The size of the plot (default is (6, 4)).
    """
    # Extract silhouette scores into a DataFrame
    df = pd.DataFrame(
        model.silhouette_scores, 
        columns=list(range(*(max(2, k_range[0] - 1), k_range[1] + 1)))  # k_range[0] is the lower bound, k_range[1] is the upper bound
    )

    # Calculate mean and standard deviation of the silhouette scores
    mean_scores = df.mean(axis=0)
    std_scores = df.std(axis=0)

    # Plotting the silhouette scores
    plt.figure(figsize=figsize)
    sns.lineplot(x=mean_scores.index, y=mean_scores, marker='o', color='cornflowerblue', label='Mean silhouette score')
    plt.fill_between(mean_scores.index, mean_scores - std_scores, mean_scores + std_scores, 
                     alpha=0.3, color='gray', label='Variance')

    # Customize plot
    plt.xlabel('Number of Clusters (k)')
    plt.ylabel('Silhouette Score')
    plt.title('Silhouette Score vs. Number of Clusters')
    plt.legend()
    plt.show()



import scanpy as sc
import matplotlib.pyplot as plt

def enrichment_score(adata_tmp, score_name, if_filtered=False):
    """
    Plot the spatial data, either the original or filtered based on the `if_filtered` parameter.
    
    Parameters:
    - adata_tmp: AnnData object, the spatial data.
    - score_name: str, the name of the score (e.g., Bcell_enrichment) to color the plot.
    - if_filtered: bool, whether to plot the filtered data (True) or the original data (False).
    """
    fig, axes = plt.subplots(figsize=(12, 6))

    if if_filtered:
        # Plot the filtered spatial data
        sc.pl.spatial(adata_tmp[isin_], color=score_name, ax=axes[1], show=False, spot_size=50)
        axes[1].set_title( score_name + "(Filtered)")
    else:
        # Plot the original spatial data
        sc.pl.spatial(adata_tmp, color=score_name, ax=axes[0], show=False, spot_size=50)
        axes[0].set_title( score_name + "(Original)")
    
    plt.show()



from joblib import Parallel, delayed
from sklearn.neighbors import KernelDensity
from collections import defaultdict
import numpy as np
from scipy.ndimage import binary_opening
import multiprocessing

def kde_filter(
    adata_tmp,
    score_name,
    high_binSize=100,
    default_thread_num=36,
    clean_mask_size=(3, 3)
):
    """
    Filter spatial coordinates based on KDE clustering.

    Args:
        adata_tmp (AnnData): Input AnnData object.
        score_name (str): Name of the score column in `.obs`.
        high_binSize (int): Grid size for KDE estimation.
        default_thread_num (int): Number of threads for parallel processing.
        clean_mask_size (Tuple[int, int]): Size of the structure for binary cleaning.

    Returns:
        adata_filtered: Filtered AnnData object.
        mask: Boolean mask of selected coordinates.
    """
    # Extract spatial coordinates and scores
    positions_original = adata_tmp.obsm['spatial']
    x = positions_original[:, 0]
    y = positions_original[:, 1]
    Z = np.array(adata_tmp.obs[score_name]).astype(np.float32)

    x_min, x_max = x.min(), x.max()
    y_min, y_max = y.min(), y.max()
    bandwidth = int(high_binSize / 3)

    points = np.vstack([x, y]).T
    xy_dict = defaultdict(list)
    for i in range(len(x)):
        Bx = (x[i] - x_min) // high_binSize * high_binSize + x_min
        By = (y[i] - y_min) // high_binSize * high_binSize + y_min
        xy_dict[(Bx, By)].append([x[i], y[i]])

    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth)
    kde.fit(points, sample_weight=Z)

    x_grid, y_grid = np.meshgrid(
        np.arange(x_min, x_max, high_binSize),
        np.arange(y_min, y_max, high_binSize)
    )
    positions = np.vstack([x_grid.ravel(), y_grid.ravel()]).T

    # Calculate KDE density
    num_cores = min(multiprocessing.cpu_count(), default_thread_num)
    density = Parallel(n_jobs=num_cores)(
        delayed(lambda pos: np.exp(kde.score_samples([pos])))(pos)
        for pos in positions
    )
    density = np.array(density).reshape(x_grid.shape)

    # Apply threshold and clean mask
    density_thresholded = np.percentile(density.ravel(), 65)
    binary_mask = ~np.isnan(density > density_thresholded)
    cleaned_mask = binary_opening(binary_mask, structure=np.ones(clean_mask_size))

    density_cleaned = np.where(cleaned_mask, density, np.nan)
    mask = ~np.isnan(density_cleaned)

    x_coords = x_grid[mask]
    y_coords = y_grid[mask]
    coords = np.vstack([x_coords, y_coords]).T

    # Map grid to original positions
    original_coords = []
    for coord in coords:
        original_coords += xy_dict[tuple(coord)]
    original_coords_set = {tuple(row) for row in original_coords}

    isin_ = [tuple(pos) in original_coords_set for pos in positions_original]
    return adata_tmp[isin_]



import scanpy as sc
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

def kde_results(adata_tmp, score_name, plot_masks=True, plot_enrichment=True, if_filtered=False, figsize=(10, 10), spot_size=50):
    """
    Plot binary_mask and mask on the first row, enrichment scores on the second row.

    Parameters
    ----------
    adata_tmp : AnnData
        AnnData object containing the data and masks.
    score_name : str
        The gene or feature name used for plotting enrichment scores.
    plot_masks : bool, optional (default=True)
        Whether to plot binary_mask and mask.
    plot_enrichment : bool, optional (default=True)
        Whether to plot the enrichment score.
    if_filtered : bool, optional (default=False)
        If True and 'filtered_coords' exists, both original and filtered enrichment scores will be plotted.
    figsize : tuple, optional (default=(10, 10))
        Size of the figure.
    spot_size : int, optional (default=50)
        Size of the points in spatial plots.
    """

    num_rows = 2 
    num_cols = 2 

    fig, axes = plt.subplots(num_rows, num_cols, figsize=figsize)

    if num_rows == 1:
        axes = [axes]  
    elif num_cols == 1:
        axes = [[ax] for ax in axes]  

    if plot_enrichment:
        if if_filtered and 'filtered_coords' in adata_tmp.obs:
            filtered_adata = adata_tmp[adata_tmp.obs['filtered_coords']]
            
            sc.pl.spatial(adata_tmp, color=score_name, ax=axes[1][0], show=False, spot_size=spot_size)
            axes[0][0].set_title("Original Enrichment Score")

            sc.pl.spatial(filtered_adata, color=score_name, ax=axes[1][1], show=False, spot_size=spot_size)
            axes[0][1].set_title("Filtered Enrichment Score")
        else:
            sc.pl.spatial(adata_tmp, color=score_name, ax=axes[1][0], show=False, spot_size=spot_size)
            axes[0][0].set_title("Enrichment Score")
            axes[0][1].axis("off")

    if plot_masks:
        if "binary_mask" not in adata_tmp.uns or "mask" not in adata_tmp.uns:
            raise ValueError("binary_mask or mask is not stored in adata_tmp.uns!")

        binary_mask = adata_tmp.uns["binary_mask"]
        mask = adata_tmp.uns["mask"]

        axes[1][0].matshow(binary_mask, cmap=ListedColormap(['silver', 'indianred']))
        axes[1][0].set_title("Binary Mask")

        axes[1][1].matshow(mask, cmap=ListedColormap(['silver', 'indianred']))
        axes[1][1].set_title("Mask")

    plt.tight_layout()
    plt.show()
